/*
 * A flexpage abstraction for handling map requests.
 *
 * Copyright (C) 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/sys/types.h>



/* A flexpage abstraction for handling map requests.

file      --------*-----------------------*-------
                  | start                 | start + length
dataspace         ----*-------------------
                      | base offset
flexpage              --------*-------
                      :       | flexpage offset
page                  :       ----    :
                      : base  : page  : base + flexpage size
memory    ------------*-------*-------------------
*/

class Flexpage
{
  /* Computed details: flexpage base address, requested page address, flexpage size. */

  l4_addr_t _base_addr;
  l4_addr_t _page_addr;
  unsigned char _order;

  /* Dataspace offset for the base of the flexpage. */

  unsigned long _base_offset;

  /* Dataspace flags. */

  unsigned long _flags;

  /* Memory region details. */

  l4_addr_t _buffer;
  unsigned long _buffer_size;

  /* Dataspace start in file and size. */

  unsigned long _data_start, _data_length;

protected:
  unsigned int _refcount;

public:
  explicit Flexpage(l4_addr_t buffer, unsigned long buffer_size)
  : _order(0), _buffer(buffer), _buffer_size(buffer_size),
    _data_start(0), _data_length(0), _refcount(0)
  {
  }

  /* Null flexpage. */

  explicit Flexpage() : _order(0), _data_length(0), _refcount(0) { }

  /* Initialisation. */

  void reset(unsigned long offset, l4_addr_t hot_spot, int flags);

  /* Reference counting. */

  unsigned int decrement() { if (_refcount) _refcount--; return _refcount; }

  void increment() { _refcount++; }

  /* Accessors. */

  l4_addr_t get_base() const { return _base_addr; }

  unsigned char get_order() const { return _order; }

  unsigned long get_size() const { return 1 << _order; }

  unsigned long get_data_start() const { return _data_start; }

  unsigned long get_data_length() const { return _data_length; }

  /* Flexpage invalidation and status. */

  void invalidate() { _order = 0; }

  bool overlap(unsigned long start, unsigned long end);

  void upgrade(unsigned long flags);

  void unmap();

  int writable();

  int valid() { return _order != 0; }

  /* Data region methods. */

  void set_data_region(unsigned long data_position,
                       unsigned long data_size);

  void set_data_size(unsigned long data_size);

  void pad();

  void clear();

  /* Comparisons. */

  bool supports(unsigned long data_position);

  bool supports_flags(unsigned long flags);

  /* Return a flexpage item. */

  l4_fpage_t get_fpage(unsigned long flags);
};
