/*
 * Accessor flexpage mapping.
 *
 * Copyright (C) 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/c/util/cap_alloc.h>
#include <l4/re/env.h>
#include <l4/sys/factory.h>
#include <l4/sys/types.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "access_map.h"

/* Obtain declarations not obtained in the access map header. */

#include "pages.h"



/* Initialise a page mapping. */

AccessMap::AccessMap()
{
  _semaphore = l4re_util_cap_alloc();

  if (l4_is_valid_cap(_semaphore))
  {
    l4_factory_create(l4re_env()->factory, L4_PROTO_SEMAPHORE, _semaphore);
    unlock_map();
  }
}

AccessMap::~AccessMap()
{
  if (l4_is_valid_cap(_semaphore))
    l4re_util_cap_free_um(_semaphore);
}

/* Find a flexpage for the given details. */

Flexpage *AccessMap::find(unsigned long data_position)
{
  lock_map();

  Flexpage *flexpage = _find(data_position);

  unlock_map();
  return flexpage;
}

Flexpage *AccessMap::_find(unsigned long data_position)
{
  Flexpage *flexpage = NULL;

  /* Iterator for accessing the mapping. */

  Flexpage_map::iterator it;

  /* Seek to the first flexpage beyond the given position. */

  it = _map.upper_bound(data_position);

  /* Obtain any previous flexpage. */

  if ((_map.size() > 0) && (it != _map.begin()))
  {
    it--;

    Flexpage *test = it->second;

    /* Test the flexpage for suitability. */

    if (test->valid() && test->supports(data_position))
      flexpage = test;
  }

  return flexpage;
}

/* Flush all allocated flexpages within a given region, writing data to
   storage. If the keep indicator is set to a true value, the flexpages will be
   retained; otherwise, they are invalidated. */

void AccessMap::flush_all(unsigned long start, unsigned long end, bool keep,
                          PagingAccessor *accessor, Pages *pages)
{
  lock_map();

  Flexpage_map::iterator it, current;

  it = _map.begin();

  while ((_map.size() != 0) && (it != _map.end()))
  {
    current = it;
    Flexpage *flexpage = current->second;
    it++;

    /* Flush valid, overlapping mapping entries. */

    if (flexpage->overlap(start, end))
    {
      accessor->flush_flexpage(flexpage, keep);

      /* If now invalid, remove from the mapping. */

      if (!flexpage->valid())
      {
        _map.erase(current);

        if (pages->reserve_flexpage(accessor, flexpage))
          pages->release_flexpage(flexpage);
      }
    }
  }

  unlock_map();
}
/* Insert the flexpage into the collection. */

void AccessMap::insert(Flexpage *flexpage)
{
  lock_map();

  _map.insert(Flexpage_map_item(flexpage->get_data_start(), flexpage));

  unlock_map();
}

/* Remove the given flexpage from the collection, flushing its data. */

bool AccessMap::remove(Flexpage *flexpage)
{
  lock_map();

  if (_find(flexpage->get_data_start()) == NULL)
  {
    unlock_map();
    return false;
  }

  _map.erase(flexpage->get_data_start());

  unlock_map();
  return true;
}

/* Purge all flexpages from the collection, flushing their data. */

void AccessMap::purge(PagingAccessor *accessor, Pages *pages)
{
  lock_map();

  Flexpage_map::iterator it;

  for (it = _map.begin(); it != _map.end(); it++)
  {
    Flexpage *flexpage = it->second;

    if (pages->reserve_flexpage(accessor, flexpage))
    {
      accessor->flush_flexpage(flexpage, false, true);
      pages->release_flexpage(flexpage);
    }
  }

  _map.clear();

  unlock_map();
}
