/*
 * File resource functionality.
 *
 * Copyright (C) 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdio.h>
#include <stdlib.h> /* posix_memalign */
#include <string.h> /* memset */

#include <ipc/server.h>
#include <systypes/stat.h>

#include "mapped_file_object_server.h"
#include "file_resource.h"
#include "resource_server.h"
#include "stat_resource.h"



/* Resource-closing methods. */

void FileResourceBase::close()
{
  /* NOTE: For debugging. */

  printf("Closing...\n");

  /* Release any locks. */

  end();

  /* Detach from the accessor, freeing it if no object is using it. */

  if (!_accessor->detach())
  {
    printf("Delete accessor: %p\n", _accessor);
    delete _accessor;
  }
}

void MappedFileResource::close()
{
  /* Flush any open pages. */

  flush_all(false);

  /* Release locks and detach the server side of the descriptor buffer. */

  FileResourceBase::close();

  /* NOTE: For debugging. */

  printf("Closed file.\n");
}



/* IPC operation methods. */

long FileResourceBase::begin(size_t *size)
{
  /* Join the queue. */

  long err = SyncSupport::begin();

  if (err)
    return err;

  _obj.size = get_size();
  *size = _obj.size;
  return L4_EOK;
}

long FileResourceBase::begin_first(size_t *size)
{
  /* Join the queue. */

  long err = SyncSupport::begin_first();

  if (err)
    return err;

  _obj.size = get_size();
  *size = _obj.size;
  return L4_EOK;
}

long FileResourceBase::flush(size_t data_start, size_t *size)
{
  size_t new_size = _obj.start_pos + data_start;

  /* Determine any new limit to the file. */

  if (new_size > _obj.size)
  {
    _obj.size = new_size;
    _accessor->set_size(new_size);
  }

  /* Flush the actual storage. */

  if (_accessor->flush())
    return -L4_EIO;

  /* Send the new size. */

  *size = _obj.size;
  return L4_EOK;
}

long FileResourceBase::fstat(size_t *size, l4_cap_idx_t *statcap)
{
  void *buffer;
  unsigned long buffer_size = l4_round_page(sizeof(sys_stat_t));

  /* NOTE: Allocation should be delegated to a page collection. */

  if (posix_memalign(&buffer, L4_PAGESIZE, buffer_size))
    return -L4_ENOMEM;

  /* Populate the allocated memory to make it available. */

  memset(buffer, 0, buffer_size);

  /* Copy the metadata into the buffer and obtain the size. */

  *size = _accessor->fstat((char *) buffer, sizeof(sys_stat_t));

  if (!*size)
    return -L4_EIO;

  /* Start a server in a new thread for the context. */

  StatResource *resource = new StatResource(buffer, buffer_size);
  ResourceServer server(resource);
  long err = server.start_thread();

  *statcap = server.config()->server;
  return err;
}


long FileResourceBase::yield(size_t *size)
{
  /* Yield to the next member of the queue. */

  SyncSupport::yield();

  _obj.size = get_size();
  *size = _obj.size;
  return L4_EOK;
}



/* Server details. */

int MappedFileResource::expected_items()
{
  return MappedFileObject_expected_items;
}

ipc_server_handler_type MappedFileResource::handler()
{
  return (ipc_server_handler_type) handle_MappedFileObject;
}

/* Memory-mapped resource operations. */

long MappedFileResource::flush(size_t data_start, size_t *size)
{
  /* Synchronise the contents of flexpages with the storage, keeping the
     flexpages. */

  flush_all(true);

  /* Perform the underlying flush operation to storage. */

  return FileResourceBase::flush(data_start, size);
}

long MappedFileResource::mmap(size_t position, size_t length, size_t *start_pos, size_t *end_pos, size_t *data_end)
{
  /* Flush and unmap all current flexpages. */

  flush_all(false);

  /* Obtain the position and length of the region to set its limits. */

  _obj.start_pos = position;
  _obj.end_pos = _obj.start_pos + length;

  /* Round the start down to the nearest page. */

  _obj.start_pos = l4_trunc_page(_obj.start_pos);

  /* Round the end up to the nearest page. */

  _obj.end_pos = l4_round_page(_obj.end_pos);

  /* Return the actual limits of the region. */

  *start_pos = _obj.start_pos;
  *end_pos = _obj.end_pos;

  /* Return the populated size of the region. */

  if (_obj.end_pos > _obj.size)
    *data_end = _obj.size - _obj.start_pos;
  else
    *data_end = _obj.end_pos - _obj.start_pos;

  return L4_EOK;
}
